﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using XorShift;

namespace MapGenerator
{
	public sealed class Division
	{
		private readonly Map map;
		private readonly Xor128 rnd;
		private readonly int left;
		private readonly int top;
		private readonly int width;
		private readonly int height;
		private readonly List<Division> partnerDivList = new List<Division>();
		private readonly List<Point> connectionList = new List<Point>();
		private readonly List<Point> doorwayList = new List<Point>();
		private readonly Dictionary<Point, Point[]> passageDic = new Dictionary<Point, Point[]>();
		private Room room;

		public int Left { get { return this.left; } }
		public int Top { get { return this.top; } }
		public int Width { get { return this.width; } }
		public int Height { get { return this.height; } }
		public int Right { get { return this.left + this.width - 1; } }
		public int Bottom { get { return this.top + this.height - 1; } }
		public Room Room { get { return this.room; } }

		public Division(Map map, Xor128 rnd, int left, int top, int width, int height)
		{
			this.map = map;
			this.rnd = rnd;
			this.left = left;
			this.top = top;
			this.width = width;
			this.height = height;
		}

		public Division[] SplitLeftAndRight(int center)
		{
			// 分割線の始点か終点につなぎ目がある場合、失敗
			if (this.connectionList.Contains(new Point(center, this.top)) || this.connectionList.Contains(new Point(center, this.Bottom)))
			{
				return null;
			}
			// 左右の幅を計算
			int leftWidth = center - this.left + 1;
			int rightWidth = this.Right - center + 1;
			// 左右に分割
			Division leftDiv = new Division(this.map, this.rnd, this.left, this.top, leftWidth, this.height);
			Division rightDiv = new Division(this.map, this.rnd, center, this.top, rightWidth, this.height);
			// つなぎ目がある場合（初回以外はあるはず）、引き継ぐ区画を判定
			foreach (Point connection in this.connectionList)
			{
				// つなぎ目が左右どちら側か判定
				Division div = connection.X < center ? leftDiv : rightDiv;
				div.connectionList.Add(connection);
				// 接続されている区画の保持するこの区画への参照を破棄し、左右の新しく接続した区画の参照を渡す
				foreach (Division connectedDiv in this.partnerDivList)
				{
					if (connectedDiv.connectionList.Contains(connection))
					{
						connectedDiv.partnerDivList.Remove(this);
						connectedDiv.partnerDivList.Add(div);
						div.partnerDivList.Add(connectedDiv);
						break;
					}
				}
			}
			// 左右をつなぐ
			Point newConnection = new Point(center, this.rnd.Next(this.top + 1, this.Bottom));
			leftDiv.connectionList.Add(newConnection);
			leftDiv.partnerDivList.Add(rightDiv);
			rightDiv.connectionList.Add(newConnection);
			rightDiv.partnerDivList.Add(leftDiv);
			// マップのマスに接点フラグを追加
			this.map[newConnection] |= Grid.Connection;
			return new Division[2] { leftDiv, rightDiv };
		}

		public Division[] SplitUpperAndUnder(int middle)
		{
			// 分割線の始点か終点につなぎ目がある場合、失敗
			if (this.connectionList.Contains(new Point(this.left, middle)) || this.connectionList.Contains(new Point(this.Right, middle)))
			{
				return null;
			}
			// 上下の高さを計算
			int upperHeight = middle - this.top + 1;
			int underHeight = this.Bottom - middle + 1;
			// 上下に分割
			Division upperDiv = new Division(this.map, this.rnd, this.left, this.top, this.width, upperHeight);
			Division underDiv = new Division(this.map, this.rnd, this.left, middle, this.width, underHeight);
			// つなぎ目がある場合（初回以外はあるはず）、引き継ぐ区画を判定
			foreach (Point connection in this.connectionList)
			{
				// つなぎ目が上下どちら側か判定
				Division div = connection.Y < middle ? upperDiv : underDiv;
				div.connectionList.Add(connection);
				// 接続されている区画の保持するこの区画への参照を破棄し、上下の新しく接続した区画の参照を渡す
				foreach (Division connectedDiv in this.partnerDivList)
				{
					if (connectedDiv.connectionList.Contains(connection))
					{
						connectedDiv.partnerDivList.Remove(this);
						connectedDiv.partnerDivList.Add(div);
						div.partnerDivList.Add(connectedDiv);
						break;
					}
				}
			}
			// 上下をつなぐ
			Point newConnection = new Point(this.rnd.Next(this.left + 1, this.Right), middle);
			upperDiv.connectionList.Add(newConnection);
			upperDiv.partnerDivList.Add(underDiv);
			underDiv.connectionList.Add(newConnection);
			underDiv.partnerDivList.Add(upperDiv);
			// マップのマスに接点フラグを追加
			this.map[newConnection] |= Grid.Connection;
			return new Division[2] { upperDiv, underDiv };
		}

		public Division[] SplitRandom(int minWidth, int minHeight)
		{
			Division[] divisions = null;
			// 先に縦の判定をするか横の判定をするかランダムに決める
			for (int i = 0, bit = this.rnd.Next(2); i < 2; ++i)
			{
				if ((i ^ bit) == 0)
				{
					// 左右に分割できるかどうか判定
					int length = (minWidth << 1) - 1;
					if (length <= this.width)
					{
						divisions = this.SplitLeftAndRight((length == this.width ? -1 : this.rnd.Next(this.width - length)) + this.left + minWidth);
						if (divisions != null)
						{
							break;
						}
					}
				}
				else
				{
					// 上下に分割できるかどうか判定
					int length = (minHeight << 1) - 1;
					if (length <= this.height)
					{
						divisions = this.SplitUpperAndUnder((length == this.height ? -1 : this.rnd.Next(this.height - length)) + this.top + minHeight);
						if (divisions != null)
						{
							break;
						}
					}
				}
			}
			return divisions;
		}

		public void CreateRoom()
		{
			int width = this.rnd.Next(this.map.MinRoomSize.Width, Math.Min(this.width - 3, this.map.MaxRoomSize.Width + 1));
			int height = this.rnd.Next(this.map.MinRoomSize.Height, Math.Min(this.height - 3, this.map.MaxRoomSize.Height + 1));
			this.room = new Room(this.left + 2 + this.rnd.Next(this.width - width - 3), this.top + 2 + this.rnd.Next(this.height - height - 3), width, height);
			// マップのマスに部屋フラグを追加
			for (int x = this.room.Left; x <= this.room.Right; ++x)
			{
				for (int y = this.room.Top; y <= this.room.Bottom; ++y)
				{
					this.map[x, y] |= Grid.Room;
				}
			}
		}

		private Point CreateTopOrBottomDoorway(int y)
		{
			int startX = this.rnd.Next(this.room.Width) + this.room.Left;
			int x = startX;
			while (true)
			{
				if (!this.doorwayList.Contains(new Point(x - 1, y)) && !this.doorwayList.Contains(new Point(x + 1, y)))
				{
					return new Point(x, y);
				}
				if (++x > this.room.Right)
				{
					x = this.room.Left;
				}
				if (x == startX)
				{
					throw new ApplicationException("通路の生成に失敗しました。");
				}
			}
		}

		private Point CreateLeftOrRightDoorway(int x)
		{
			int startY = this.rnd.Next(this.room.Height) + this.room.Top;
			int y = startY;
			while (true)
			{
				if (!this.doorwayList.Contains(new Point(x, y - 1)) && !this.doorwayList.Contains(new Point(x, y + 1)))
				{
					return new Point(x, y);
				}
				if (++y > this.room.Bottom)
				{
					y = this.room.Top;
				}
				if (y == startY)
				{
					throw new ApplicationException("通路の生成に失敗しました。");
				}
			}
		}

		private void CreatePassage(Point connection)
		{
			// 同じ Grid に何度でも Doorway を作れるようにすると、通路生成に失敗しないはず
			List<Point> route = new List<Point>();
			Point doorway;
			int r = this.rnd.Next(64);
			if (r < 48)
			{
				// 75% の確率で、接点に対して正面の辺に Doorway を作る
				if (connection.X == this.left)
				{
					if ((r & 2) != 0 && connection.Y >= this.room.Top && connection.Y <= this.room.Bottom && (this.map[this.room.Left - 1, connection.Y - 1] & Grid.Doorway) == 0 && (this.map[this.room.Left - 1, connection.Y + 1] & Grid.Doorway) == 0)
					{
						// 50% の確率で、さらに、接点の Y 座標が部屋の上と下の Y 座標に収まり、Doorway 予定座標の上下に Doorway が無い場合、接点に対して一直線に通路を引く
						doorway = new Point(this.room.Left - 1, connection.Y);
						route.AddRange(new Point[2] { doorway, new Point(this.left, doorway.Y) });
					}
					else
					{
						doorway = this.CreateLeftOrRightDoorway(this.room.Left - 1);
						route.AddRange(new Point[2] { doorway, new Point(connection.X, doorway.Y) });
					}
				}
				else if (connection.Y == this.top)
				{
					if ((r & 2) != 0 && connection.X >= this.room.Left && connection.X <= this.room.Right && (this.map[connection.X - 1, this.room.Top - 1] & Grid.Doorway) == 0 && (this.map[connection.X + 1, this.room.Top - 1] & Grid.Doorway) == 0)
					{
						// 50% の確率で、さらに、接点の X 座標が部屋の左と右の X 座標に収まり、Doorway 予定座標の左右に Doorway が無い場合、接点に対して一直線に通路を引く
						doorway = new Point(connection.X, this.room.Top - 1);
						route.AddRange(new Point[2] { doorway, new Point(doorway.X, this.top) });
					}
					else
					{
						doorway = this.CreateTopOrBottomDoorway(this.room.Top - 1);
						route.AddRange(new Point[2] { doorway, new Point(doorway.X, connection.Y) });
					}
				}
				else if (connection.X == this.Right)
				{
					if ((r & 2) != 0 && connection.Y >= this.room.Top && connection.Y <= this.room.Bottom && (this.map[this.room.Right + 1, connection.Y - 1] & Grid.Doorway) == 0 && (this.map[this.room.Right + 1, connection.Y + 1] & Grid.Doorway) == 0)
					{
						// 50% の確率で、さらに、接点の Y 座標が部屋の上と下の Y 座標に収まり、Doorway 予定座標の上下に Doorway が無い場合、接点に対して一直線に通路を引く
						doorway = new Point(this.room.Right + 1, connection.Y);
						route.AddRange(new Point[2] { doorway, new Point(this.Right, doorway.Y) });
					}
					else
					{
						doorway = this.CreateLeftOrRightDoorway(this.room.Right + 1);
						route.AddRange(new Point[2] { doorway, new Point(connection.X, doorway.Y) });
					}
				}
				else
				{
					if ((r & 2) != 0 && connection.X >= this.room.Left && connection.X <= this.room.Right && (this.map[connection.X - 1, this.room.Bottom + 1] & Grid.Doorway) == 0 && (this.map[connection.X + 1, this.room.Bottom + 1] & Grid.Doorway) == 0)
					{
						// 50% の確率で、さらに、接点の X 座標が部屋の左と右の X 座標に収まり、Doorway 予定座標の左右に Doorway が無い場合、接点に対して一直線に通路を引く
						doorway = new Point(connection.X, this.room.Bottom + 1);
						route.AddRange(new Point[2] { doorway, new Point(doorway.X, this.Bottom) });
					}
					else
					{
						doorway = this.CreateTopOrBottomDoorway(this.room.Bottom + 1);
						route.AddRange(new Point[2] { doorway, new Point(doorway.X, connection.Y) });
					}
				}
			}
			else if (r < 55)
			{
				// 10.9375% の確率で、接点に対して左側の辺に Doorway を作る
				if (connection.X == this.left)
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Bottom + 1);
					route.AddRange(new Point[3] { doorway, new Point(doorway.X, this.Bottom), new Point(connection.X, this.Bottom) });
				}
				else if (connection.Y == this.top)
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Left - 1);
					route.AddRange(new Point[3] { doorway, new Point(this.left, doorway.Y), new Point(this.left, connection.Y) });
				}
				else if (connection.X == this.Right)
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Top - 1);
					route.AddRange(new Point[3] { doorway, new Point(doorway.X, this.top), new Point(connection.X, this.top) });
				}
				else
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Right + 1);
					route.AddRange(new Point[3] { doorway, new Point(this.Right, doorway.Y), new Point(this.Right, connection.Y) });
				}
			}
			else if (r < 62)
			{
				// 10.9375% の確率で、接点に対して右側の辺に Doorway を作る
				if (connection.X == this.left)
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Top - 1);
					route.AddRange(new Point[3] { doorway, new Point(doorway.X, this.top), new Point(connection.X, this.top) });
				}
				else if (connection.Y == this.top)
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Right + 1);
					route.AddRange(new Point[3] { doorway, new Point(this.Right, doorway.Y), new Point(this.Right, connection.Y) });
				}
				else if (connection.X == this.Right)
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Bottom + 1);
					route.AddRange(new Point[3] { doorway, new Point(doorway.X, this.Bottom), new Point(connection.X, this.Bottom) });
				}
				else
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Left - 1);
					route.AddRange(new Point[3] { doorway, new Point(this.left, doorway.Y), new Point(this.left, connection.Y) });
				}
			}
			else
			{
				// 3.125% の確率で、接点に対して反対側の辺に Doorway を作る
				if (connection.X == this.left)
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Right + 1);
					int y = (r & 1) != 0 ? this.top : this.Bottom;
					route.AddRange(new Point[4] { doorway, new Point(this.Right, doorway.Y), new Point(this.Right, y), new Point(connection.X, y) });
				}
				else if (connection.Y == this.top)
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Bottom + 1);
					int x = (r & 1) != 0 ? this.left : this.Right;
					route.AddRange(new Point[4] { doorway, new Point(doorway.X, this.Bottom), new Point(x, this.Bottom), new Point(x, connection.Y) });
				}
				else if (connection.X == this.Right)
				{
					doorway = this.CreateLeftOrRightDoorway(this.room.Left - 1);
					int y = (r & 1) != 0 ? this.Bottom : this.top;
					route.AddRange(new Point[4] { doorway, new Point(this.left, doorway.Y), new Point(this.left, y), new Point(connection.X, y) });
				}
				else
				{
					doorway = this.CreateTopOrBottomDoorway(this.room.Top - 1);
					int x = (r & 1) != 0 ? this.Right : this.left;
					route.AddRange(new Point[4] { doorway, new Point(doorway.X, this.top), new Point(x, this.top), new Point(x, connection.Y) });
				}
			}
			if (!this.doorwayList.Contains(doorway))
			{
				this.doorwayList.Add(doorway);
			}
			foreach (Division div in this.partnerDivList)
			{
				if (div.passageDic.ContainsKey(connection))
				{
					{
						Point last = div.passageDic[connection].Last();
						if (route.Last() != last)
						{
							route.Add(last);
						}
					}
					break;
				}
			}
			this.passageDic.Add(connection, route.ToArray());
			// マップのマスにドア、通路フラグを追加
			this.map[route.First()] |= Grid.Doorway;
			for (int i = 1; i < route.Count; ++i)
			{
				if (route[i - 1].X == route[i].X)
				{
					int x = route[i].X;
					for (int y = Math.Min(route[i - 1].Y, route[i].Y); y <= Math.Max(route[i - 1].Y, route[i].Y); ++y)
					{
						this.map[x, y] |= Grid.Passage;
					}
				}
				else
				{
					int y = route[i].Y;
					for (int x = Math.Min(route[i - 1].X, route[i].X); x <= Math.Max(route[i - 1].X, route[i].X); ++x)
					{
						this.map[x, y] |= Grid.Passage;
					}
				}
			}
		}

		public void CreatePassages()
		{
			foreach (Point connection in this.connectionList)
			{
				this.CreatePassage(connection);
			}
		}

		private bool PointOnDivisionFrame(Point point)
		{
			return ((point.X == this.left || point.X == this.Right) && point.Y > this.top && point.Y < this.Bottom) || (point.X > this.left && point.X < this.Right && (point.Y == this.top || point.Y == this.Bottom));
		}

		public void ConnectRandom()
		{
			Point connection;
			Point sideA;
			Point sideB;
			switch (this.rnd.Next(4))
			{
				default:
				case 0:	// Left
					if (this.left <= this.map.LeftSpace)
					{
						return;
					}
					connection = new Point(this.left, this.rnd.Next(this.top + 1, this.Bottom));
					sideA = new Point(this.left, connection.Y - 1);
					sideB = new Point(this.left, connection.Y + 1);
					break;
				case 1:	// Top
					if (this.top <= this.map.TopSpace)
					{
						return;
					}
					connection = new Point(this.rnd.Next(this.left + 1, this.Right), this.top);
					sideA = new Point(connection.X - 1, this.top);
					sideB = new Point(connection.X + 1, this.top);
					break;
				case 2:	// Right
					if (this.Right >= this.map.Right - (this.map.RightSpace << 1))
					{
						return;
					}
					connection = new Point(this.Right, this.rnd.Next(this.top + 1, this.Bottom));
					sideA = new Point(this.Right, connection.Y - 1);
					sideB = new Point(this.Right, connection.Y + 1);
					break;
				case 3:	// Bottom
					if (this.Bottom >= this.map.Bottom - (this.map.BottomSpace << 1))
					{
						return;
					}
					connection = new Point(this.rnd.Next(this.left + 1, this.Right), this.Bottom);
					sideA = new Point(connection.X - 1, this.Bottom);
					sideB = new Point(connection.X + 1, this.Bottom);
					break;
			}
			if (this.connectionList.Contains(connection) || (this.map[sideA] & Grid.Connection) != 0 || (this.map[sideB] & Grid.Connection) != 0)
			{
				return;
			}
			foreach (Division div in this.partnerDivList)
			{
				if (div.PointOnDivisionFrame(connection))
				{
					this.connectionList.Add(connection);
					div.connectionList.Add(connection);
					this.map[connection] |= Grid.Connection;
					return;
				}
			}
		}
	}
}
